/****************************************************************************/
/*                                                                          */
/* These routines were written by Charles Henrich                           */
/*                                                                          */
/* Copyright (c) 1993 Michigan State University Board of Trustee's          */
/*                                                                          */
/****************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#ifdef _AIX
#include <sys/select.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <sys/time.h>

#ifdef _AIX
#include <sys/ioctl.h>
#include <sys/devinfo.h>
#include <sys/gdlextcb.h>
#else
#include <fcntl.h>
#endif

#include <sys/file.h>
#include <sys/errno.h>
#include <errno.h>

#include <sys/param.h>

#include <signal.h>

#include "st.h"
#include "stP.h"

#ifdef DES
#include "des.h"
#endif

 /**************************************/
 /*                                    */
 /* Public info-sharing structure arr. */
 /*                                    */
 /**************************************/

SOCK *callback_info[FD_SETSIZE];

 /**************************************/
 /*                                    */
 /* Private global variables used by   */
 /* the tcp server routines, internal  */
 /* use only.                          */
 /*                                    */
 /**************************************/

static int ready_descriptors[FD_SETSIZE];   /* Descriptors with data in them*/
static int all_descriptors[FD_SETSIZE];     /* List of all connected desc.  */
static int local_socket = -1;               /* Local socket descriptor      */
static int nfds;                            /* High+1 descriptor #          */
static fd_set readfds;                      /* Socket bit sets for select   */
static struct sockaddr_in local_sin;        /* Local Socket address         */
struct sockaddr_in dest_sin;                /* Destination socket address   */
static struct timeval timer;                /* Timeout struct for select    */
static CBACK_FUNC callback_list[NUM_CBACKS];/* Pointers to callback funcs   */
static int timeout = 60;                    /* Default select timeout value */
static int num_socket = 0;                  /* Number of outgoing sockets   */
static int local_len = sizeof(local_sin);

char *G_errortable[] =
    {
    "Unable to instantiate a local socket",
    "Unable to bind socket",
    "Unable to listen on port",
    "Network select failed",
    "The number of active descriptors isnt what we think it should be",
    "Unable to reset descriptor list",
    "Unknown callback",
    "Only received a partial string",
    "Unknown Hostname",
    "Cant get a socket",
    "Invalid socket, (we've never seen it before)",
    "Timeout occured",
    "We have completed all desired iterations",
    "Unable to set socket option",
    "Unable to set DES key"
    };

#ifdef DES
int G_encrypt = FALSE;
des_cblock G_deskey;
des_key_schedule G_dessched;
#endif

/*--------------------------------------------------------------------------*/
/* PUBLIC ROUTINES                                                          */
/*--------------------------------------------------------------------------*/

/****************************************************************************/
/* Function: st_error                   Author: Charles Henrich             */
/*                                      Date  : October 25th 1993           */
/*                                                                          */
/* Calling Parameters: int error, char *string, int strln                   */
/*                                                                          */
/* Return codes: NONE                                                       */
/*                                                                          */
/* Purpose: This function places an english representation of the error     */
/*          passed in in the given string.                                  */
/*                                                                          */
/****************************************************************************/

char *st_error(int error)
{
static char errbuf[100];

if(error < 0) error *= -1;                    /* We want positive values :) */
error--;                                      /* Our error codes start at 0 */

if(error >= ERROR_TABLE_SIZE) strcpy(errbuf, "Unknown error!");
else                          strcpy(errbuf, G_errortable[error]);

return errbuf;
}

/****************************************************************************/
/* Function: st_close                   Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: SOCK *entry                                          */
/*                                                                          */
/* Return codes: NONE                                                       */
/*                                                                          */
/* Purpose: This function closes the connection referenced by the SOCK      */
/*          structure                                                       */
/*                                                                          */
/****************************************************************************/

void st_close(SOCK *entry)
{
#ifdef DEBUG
printf("ST:st_close\n");
#endif

if(entry == NULL) return;
close_connection(entry->descriptor);
}

/****************************************************************************/
/* Function: st_reset_callbacks         Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: NONE                                                 */
/*                                                                          */
/* Return codes: SUCCESS                                                    */
/*                                                                          */
/* Purpose: This function clears all callback's that have been set          */
/*                                                                          */
/****************************************************************************/

int st_reset_callbacks(void)
{
int count;

#ifdef DEBUG
printf("ST:Entered socket_reset_callbacks\n");
#endif

for(count=0;count<NUM_CBACKS;count++)
    {
    callback_list[count] = NULL;
    }

return SUCCESS;
}

/****************************************************************************/
/* Function: st_write                   Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: SOCK *entry, char *fmt, ...                          */
/*                                                                          */
/* Return codes: SUCCESS or FAILURE or MASSIVE_FAILURE                      */
/*                                                                          */
/* Purpose: This routine will write a printf() style formatted string to    */
/*          a socket.                                                       */
/*                                                                          */
/****************************************************************************/

int st_write(SOCK *entry, char *fmt, ...)
{
char buffer[BUFSIZ];
va_list ap;

#ifdef DEBUG
printf("ST:Writing %s to %d\n",string, entry->descriptor);
#endif

if(entry == NULL) return FAILURE;

va_start(ap, fmt);
vsprintf(buffer, fmt, ap);
va_end(ap);

return sock_write(entry->descriptor, buffer);
}

/****************************************************************************/
/* Function: st_set_timeout             Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: int timeout (timeout is number of seconds)           */
/*                                                                          */
/* Return codes: SUCCESS                                                    */
/*                                                                          */
/* Purpose: This function sets the timeout value of the select statements   */
/*                                                                          */
/****************************************************************************/

int st_set_timeout(int timeout)
{
timer.tv_sec = timeout;
timer.tv_usec = 0;

return SUCCESS;
}

/****************************************************************************/
/* Function: st_remove_callback         Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: int callback                                         */
/*                                                                          */
/* Return codes: SUCCESS or INVALID_CALLBACK                                */
/*                                                                          */
/* Purpose: This function clears a specific callback                        */
/*                                                                          */
/****************************************************************************/

int st_remove_callback(int callback)
{
#ifdef DEBUG
printf("ST:Entered socket_remove_callback\n");
#endif

if(callback >= NUM_CBACKS) return INVALID_CALLBACK;

callback_list[callback] = NULL;

return SUCCESS;
}

/****************************************************************************/
/* Function: st_add_callback            Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: The callback #, and a pointer to a function          */
/*                                                                          */
/* Return codes: SUCCESS or INVALID_CALLBACK                                */
/*                                                                          */
/* Purpose: This function activate's a callback                             */
/*                                                                          */
/****************************************************************************/

int st_add_callback(int callback, void function(SOCK *))
{
#ifdef DEBUG
printf("ST:Entered socket_add_callback\n");
#endif

if(callback >= NUM_CBACKS) return INVALID_CALLBACK;

callback_list[callback] = function;

return SUCCESS;
}

/****************************************************************************/
/* Function: st_init                    Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: DES String                                           */
/*                                                                          */
/* Return codes: SUCCESS                                                    */
/*                                                                          */
/* Purpose: This function initializes several things used by the st.        */
/*          This function must be called before any other st routine.       */
/*                                                                          */
/****************************************************************************/

int st_init(char *desstr)
{
int count;                                      /* Generic counter variable */

#ifdef DES
int ret;
#endif

#ifdef DEBUG
printf("ST:Socket init\n");
#endif

signal(SIGPIPE, SIG_IGN);                            /* Ignore broken pipes */

st_reset_callbacks();

for(count=0;count<FD_SETSIZE;count++)
    {
    all_descriptors[count]=0;
    ready_descriptors[count]=0;
    }

FD_ZERO(&readfds);                    /* Initialize descriptor sets to zero */

timer.tv_sec = timeout;
timer.tv_usec = 0;

#ifdef DES
if(desstr != NULL)
    {
    des_string_to_key(desstr, &G_deskey);

    ret=des_set_key(&G_deskey, G_dessched);

    if(ret == -1 || ret == -2) return DES_SET_KEY_FAILURE;

    G_encrypt = TRUE;
    }
else
    {
    G_encrypt = FALSE;
    }

#endif

return SUCCESS;
}

/****************************************************************************/
/* Function: st_connect                 Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: (hostname/ip), port number                           */
/*                                                                          */
/* Return codes: SOCK structure or NULL;                                    */
/*                                                                          */
/* Purpose: This function establishes a connection with a remote system.    */
/*                                                                          */
/****************************************************************************/

SOCK *st_connect(char *host, int port)
{
struct hostent *hs;
int new_descriptor;
int dotquad=1;
int count,ret;
SOCK *entry;

#ifdef DEBUG
printf("ST:Inside st_connect\n");
#endif

new_descriptor = socket(PF_INET, SOCK_STREAM, 0);

if(new_descriptor <0) return NULL;

callback_info[new_descriptor] = (SOCK *)calloc(sizeof(SOCK), sizeof(char));

entry=callback_info[new_descriptor];

#ifdef DEBUG
printf("ST:st_connect:Entry is %p\n",entry);
#endif

entry->descriptor=new_descriptor;

for(count=0;host[count];count++)
    {
    if(isalpha(host[count])) dotquad = 0;
    }

if(dotquad == 0)
    {
    hs=gethostbyname(host);

    if(hs == NULL) 
        {
        free(entry);
        close(new_descriptor);
        callback_info[new_descriptor] = NULL;
        return NULL;
        }

    memcpy(&dest_sin.sin_addr, hs->h_addr_list[0], hs->h_length);
    }
else
    {
    dest_sin.sin_addr.s_addr = inet_addr(host);
    }

dest_sin.sin_family = AF_INET;
dest_sin.sin_port = htons(port);

ret=connect(entry->descriptor, (struct sockaddr *) &dest_sin,
            sizeof(dest_sin));

#ifdef DEBUG
printf("ST:Connect returns %d\n",ret);
printf("ST:Errrno %d\n",errno);
printf("ST:Destip %s   Port %d\n",host,port);
#endif

if(ret == 0)
    {
    if(entry->descriptor >= nfds) nfds=entry->descriptor + 1;
    all_descriptors[entry->descriptor] = 1;
    FD_SET(entry->descriptor, &readfds);

    num_socket++;
    return entry;
    }

#ifdef DEBUG
perror("ST:Connect");
#endif

close(entry->descriptor);
callback_info[entry->descriptor] = NULL;
free(entry);

return NULL;
}

/****************************************************************************/
/* Function: st_listen                  Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: port #                                               */
/*                                                                          */
/* Return codes: LOCAL_SOCKET_ERROR or BIND_ERROR or LISTEN_ERROR or        */
/*               SOCKOPT_ERROR or SUCCESS                                   */
/*                                                                          */
/* Purpose: This function opens up a local port and begins to listen on it  */
/*          for incoming connections.                                       */
/*                                                                          */
/* Change History:                                                          */
/*                                                                          */
/* Charles Henrich  Now set SO_LINGER to linger for 30 seconds before       */
/* June 30th, 1993  giving up on the close.                                 */
/*                                                                          */
/****************************************************************************/

int st_listen(int port)
{
int ret;                                        /* Generic return variable  */
int tmp;                                        /* Temp int variable        */
struct linger linger;                           /* Linger options           */

#ifdef DEBUG
printf("ST:St_listen\n");
#endif

local_socket = socket(PF_INET, SOCK_STREAM, 0);       /* Setup local socket */

if(local_socket < 0)
    {
#ifdef DEBUG
perror("ST:Local Socket");
#endif

    return LOCAL_SOCKET_ERROR;
    }

linger.l_onoff = 1;
linger.l_linger = 30;
ret=setsockopt(local_socket, SOL_SOCKET, SO_LINGER,
               (char *)&linger, sizeof(linger));

if(ret < 0)
    {
#ifdef DEBUG
perror("ST:Local Socket, SO_LINGER setsockopt");
#endif
    return SOCKOPT_ERROR;
    }

tmp=1;
ret=setsockopt(local_socket, SOL_SOCKET, SO_REUSEADDR,
               (char *)&tmp, sizeof(tmp));

if(ret < 0)
    {
#ifdef DEBUG
perror("ST:Local Socket, SO_REUSEADDR setsockopt");
#endif
    return SOCKOPT_ERROR;
    }

#ifdef _AIX
tmp=1;
ret=ioctl(local_socket, FIONBIO, &tmp);         /* Make socket non-blocking */

#ifdef DEBUG
if(ret != 0) perror("ST:Ioctl");      /* Unable to make socket non-blocking */
#endif

#else /* _AIX */
ret=fcntl(local_socket, F_SETFL, FNDELAY);      /* Make socket non-blocking */

#ifdef DEBUG
if(ret < 0) perror("ST:Fcntl");       /* Unable to make socket non-blocking */
#endif

#endif /* _AIX */

nfds=local_socket+1;                                 /* Set high socket num */

 /**************************************/
 /*                                    */
 /* Add the local_socket to the list   */
 /* of active descriptors              */
 /*                                    */
 /**************************************/

all_descriptors[local_socket] = 1;

local_sin.sin_family = AF_INET;                  /* Setup local port/socket */
local_sin.sin_addr.s_addr = INADDR_ANY;
local_sin.sin_port = htons(port);

 /**************************************/
 /*                                    */
 /* Attach name to socket descriptor   */
 /*                                    */
 /**************************************/

ret = bind(local_socket, (struct sockaddr *) &local_sin, local_len);

if(ret == -1)
    {

#ifdef DEBUG
perror("ST:Bind");
#endif

    return BIND_ERROR;
    }

 /**************************************/
 /*                                    */
 /* Add the local_socket to the        */
 /* descriptor set that select watches */
 /*                                    */
 /**************************************/

FD_SET(local_socket, &readfds);

callback_info[local_socket] = (SOCK *)calloc(sizeof(SOCK),sizeof(char));

ret=listen(local_socket, 5);                         /* Set up socket queue */

if(ret == -1)
    {

#ifdef DEBUG
perror("ST:Listen");
#endif

    return LISTEN_ERROR;
    }

return SUCCESS;
}

/****************************************************************************/
/* Function: st_read                    Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: socket struct to read from, and timeout (in seconds) */
/*                                                                          */
/* Return codes: SUCCESS or INVALID_SOCK or SELECT_ERROR or FAILURE or      */
/*               TIMEOUT or MASSIVE_FAILURE                                 */
/*                                                                          */
/* Purpose: This function will wait a maximum of timeout seconds for data   */
/*          to arrive on the socket, and copy that data into the socket     */
/*          buffer.  The function returns on error or timeout or when a     */
/*          full line of text (\n terminated) has been read.                */
/*                                                                          */
/* Change History:                                                          */
/*                                                                          */
/* Charles Henrich  Wasnt properly handling a MASSIVE_FAILURE from          */
/* June 30th, 1993  sock_read, st_read now returns whatever sock_read does  */
/*                  unless it returns a PARTIAL_STRING, in which case we    */
/*                  go back to the top of the timer loop.                   */
/*                                                                          */
/****************************************************************************/

int st_read(SOCK *entry, int timeout)
{
int ret;                                        /* Generic return variable  */
long starttime;
long endtime;
static struct timeval read_timer;              /* Timeout struct for select */
static fd_set local_readfds;                  /* Socket bit sets for select */

#ifdef DEBUG
printf("ST:Entered st_read\n");
#endif

if(all_descriptors[entry->descriptor] != 1) return INVALID_SOCK;

read_timer.tv_usec = 0;

time(&starttime);
endtime = starttime;

entry->buffer[0] = '\0';

while((endtime - starttime) < timeout)
    {

#ifdef DEBUG
printf("ST:Top of st_read loop\n");
#endif

    read_timer.tv_sec = timeout - (endtime - starttime);

    FD_ZERO(&local_readfds);               /* Zero the local descriptor set */

    FD_SET(entry->descriptor, &local_readfds);  /* Add me to descriptor set */

#ifdef DEBUG
printf("ST:About to select..\n");
#endif

    ret = select(nfds, &local_readfds, (fd_set *)0, (fd_set *)0, &read_timer);

#ifdef DEBUG
printf("ST:Select returns %d\n",ret);
#endif

    if(ret == -1)                                /* Error occured in select */
        {
        if(errno == EINTR)                         /* Its nothing, continue */
            {
            time(&endtime);
            continue;
            }

        return SELECT_ERROR;
        }

    if(ret == 0) return TIMEOUT;

#ifdef DEBUG
printf("ST:Sock reading..\n");
#endif

    ret=sock_read(entry->descriptor, entry->buffer, BUF_SIZE);

#ifdef DEBUG
printf("ST:Back from sock_read, returned %d\n",ret);
#endif

    if(ret == PARTIAL_STRING)                            /* Need more data! */
        {
        time(&endtime);

        continue;
        }

 /**************************************/
 /*                                    */
 /* If the read didnt return a partial */
 /* string, we dont need to do         */
 /* anything more, just return what    */
 /* sock_read did.                     */
 /*                                    */
 /**************************************/

    return ret;
    }

#ifdef DEBUG
printf("Returning a timeout here\n");
#endif

return TIMEOUT;
}

/****************************************************************************/
/* Function: st_mainloop                Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters:  Number of iterations or 0 for endless.              */
/*                                                                          */
/* Return codes: SELECT_ERROR or NUMBER_OF_DESCRIPTORS_ERROR or             */
/*               RESET_DESCRIPTORS_ERROR or ITERATIONS_COMPLETE             */
/*                                                                          */
/* Purpose: This function will wait for incoming data on all currently      */
/*          open sockets, and call the appropriate functions to handle said */
/*          data.  There are two modes of operation here, for the first you */
/*          call this function with 0 as the iteration.  Execution will     */
/*          never leave this function unless something really bad happens.  */
/*          This is the most common mode of execution.  For the other mode  */
/*          you call this function with a number of iterations, that is how */
/*          many times you want to loop through this structure checking for */
/*          data and handling it.  At the end of the iterations control is  */
/*          returned to the caller.                                         */
/*                                                                          */
/* Change History:                                                          */
/*                                                                          */
/* Charles Henrich  We now check to see if the descriptor in                */
/* June 30th, 1993  ready_descriptors is -1, and if so, we dont do a thing  */
/*                  with it.                                                */
/*                                                                          */
/****************************************************************************/

int st_mainloop(int iterations)
{
int count;                                      /* Generic counter variable */
int ret;                                        /* Generic return variable  */
int ret1;                                       /* Generic return variable  */

#ifdef DEBUG
printf("ST:Entered st_mainloop\n");
#endif
while(1)                                 /* Loop forever, watching for data */
    {
#ifdef DEBUG
printf("ST:Top of select loop\n");
#endif

 /**************************************/
 /*                                    */
 /* Watch for incoming data, or a      */
 /* timeout condition                  */
 /*                                    */
 /**************************************/

    do_callback(st_PRE_SELECT, local_socket);

    ret = select(nfds, &readfds, (fd_set *)0, (fd_set *)0, &timer);

    if(ret == -1)                                /* Error occured in select */
        {

#ifdef DEBUG
perror("ST:Select");
#endif

        if(errno == EINTR) continue;

        return SELECT_ERROR;
        }

    if(ret == 0)                                        /* Select Timed Out */
        {
        do_callback(st_INPUT_TIMEOUT,local_socket);
        }

#ifdef DEBUG
printf("ST:Calling get_Ready_descriptors\n");
#endif

    ret1=get_ready_descriptors(&readfds, ready_descriptors);

#ifdef DEBUG
printf("ST:back from get_ready\n");
#endif

 /**************************************/
 /*                                    */
 /* get_ready_descriptors should have  */
 /* processed the same number of       */
 /* descriptors as select!             */
 /*                                    */
 /**************************************/

    if(ret1 != ret)
        {
#ifdef DEBUG
printf("ST:Number of descriptors mismatch!\n");
#endif
        return NUMBER_OF_DESCRIPTORS_ERROR;
        }

    for(count=0;count<ret1;count++)       /* Loop through ready descriptors */
        {

 /**************************************/
 /*                                    */
 /* If the socket that is ready is the */
 /* local_socket it means we have to   */
 /* start a new connection here.       */
 /*                                    */
 /**************************************/

        if(ready_descriptors[count] == local_socket)
            {
            ret=initialize_new_connection(local_socket);

#ifdef DEBUG

if(ret == FAILURE)
    {
    printf("ST: Unable to initialize new connection\n");
    }

#endif

            continue;
            }

#ifdef DEBUG
printf("ST:Calling input handler\n");
#endif

        if(ready_descriptors[count] == -1) continue;    /* Dont process us  */
                                                        /* if we arent      */
                                                        /* valid..          */

        ret=input_handler(ready_descriptors[count]);    /* Read and process */
                                                                   /* input */

#ifdef DEBUG
printf("ST:Input handler returned %d\n",ret);
#endif

        if(ret == FAILURE) continue;                  /* Input read failed! */

        if(ret == MASSIVE_FAILURE)        /* Something REALLY evil happened */
            {                                      /* close the connection! */
            close_connection(ready_descriptors[count]);
            continue;
            }
        }

    ret=reset_input_descriptors();      /* Reset the select descriptor mask */

    if(ret == FAILURE)                                      /* Reset failed */
        {
        return RESET_DESCRIPTORS_ERROR;
        }

    if(iterations == 1)
        {
        return SUCCESS;
        }
    else
        {
        if(iterations != 0) iterations--;
        }
    }                                                    /* End of while(1) */
}

/*--------------------------------------------------------------------------*/
/* PRIVATE ROUTINES                                                         */
/*--------------------------------------------------------------------------*/

/****************************************************************************/
/* Function: get_ready_descriptors      Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: the read descriptor set and an array that will be    */
/*                          used to hold the descriptor #'s that are ready  */
/*                          to be processed                                 */
/*                                                                          */
/* Return codes: the number of descriptors ready to be processed;           */
/*                                                                          */
/* Purpose: This function will fill an array with the descriptor numbers    */
/*          that are ready to be processed                                  */
/*                                                                          */
/****************************************************************************/

static int get_ready_descriptors(fd_set *readfds, int ready_descriptors[])
{
int count;
int num_descriptors;

#ifdef DEBUG
printf("ST:Get_ready_descriptors\n");
#endif

num_descriptors=0;

for(count=0;count<nfds;count++) ready_descriptors[count] = 0;

/* Loop through all possible descriptors */

for(count=0;count<nfds;count++)
    {

 /**************************************/
 /*                                    */
 /* Check to see if the descriptor has */
 /* something waiting on it.           */
 /*                                    */
 /**************************************/

    if(FD_ISSET(count, readfds) != 0)
        {

 /**************************************/
 /*                                    */
 /* If there is something waiting, add */
 /* the descriptor to the ready        */
 /* descriptors list.                  */
 /*                                    */
 /**************************************/

        ready_descriptors[num_descriptors] = count;
        num_descriptors++;
        }
    }

#ifdef DEBUG
printf("ST:Exiting get_ready_descriptors\n");
#endif

return num_descriptors;
}

/****************************************************************************/
/* Function: reset_input_descriptors    Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: NONE                                                 */
/*                                                                          */
/* Return codes: SUCCESS                                                    */
/*                                                                          */
/* Purpose: This function will reset all the descriptors that should be     */
/*          monitored for activity                                          */
/*                                                                          */
/****************************************************************************/

static int reset_input_descriptors(void)
{
int count;

#ifdef DEBUG
printf("ST:Resetting input desc.\n");
#endif

FD_ZERO(&readfds);                       /* Zero the select descriptor mask */

 /**************************************/
 /*                                    */
 /* Loop through all possible          */
 /* descriptors, finding those that    */
 /* are active, if they are, reset     */
 /* them in the select descriptor      */
 /* mask                               */
 /*                                    */
 /**************************************/

for(count=0;count<nfds;count++)
    {
    if(all_descriptors[count] == 1) FD_SET(count, &readfds);
    }

return SUCCESS;
}

/****************************************************************************/
/* Function: initialize_new_connection  Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: descriptor to be initialized                         */
/*                                                                          */
/* Return codes: SUCCESS or FAILURE                                         */
/*                                                                          */
/* Purpose: This function accepts new connections, add's them to the global */
/*          active descriptor list, allocate's a socket structure for it    */
/*          and fills in the relevant information (such as hostname)        */
/*                                                                          */
/****************************************************************************/

static int initialize_new_connection(int descriptor)
{
int new_descriptor;
struct hostent *hs;
struct sockaddr_in remote_addr;                    /* Remote Socket Address */
int remote_len = sizeof(remote_addr);

#ifdef DEBUG
printf("ST:Initialize new connection\n");
#endif

 /**************************************/
 /*                                    */
 /* Accept a new connection on the     */
 /* local socket.                      */
 /*                                    */
 /**************************************/

new_descriptor = accept(local_socket, (struct sockaddr *) &remote_addr,
                     &remote_len);

if(new_descriptor == -1) return FAILURE;        /* Nothing there to accept! */

 /**************************************/
 /*                                    */
 /* Make sure nfds is always > than    */
 /* the largest descriptor number      */
 /*                                    */
 /**************************************/

if(new_descriptor >= nfds) nfds=new_descriptor+1;

 /**************************************/
 /*                                    */
 /* Add the descriptor to the active   */
 /* connections list                   */
 /*                                    */
 /**************************************/

all_descriptors[new_descriptor] = 1;

FD_SET(new_descriptor, &readfds);          /* Add descriptor to select list */

callback_info[new_descriptor] = (SOCK *)calloc(sizeof(SOCK),sizeof(char));

hs=gethostbyaddr((char *)&remote_addr.sin_addr,
                 sizeof(remote_addr.sin_addr), AF_INET);

if(hs == NULL)
    {
#if defined(NeXT) || defined(sgi) || defined(__hpux) || defined (SOLARIS) || defined(__FreeBSD__) || defined(__NetBSD__)

    strncpy(callback_info[new_descriptor]->hostname,
            inet_ntoa(remote_addr.sin_addr),MAXHOSTNAMELEN);

#else

    strncpy(callback_info[new_descriptor]->hostname,
            inet_ntoa(&(remote_addr.sin_addr)),MAXHOSTNAMELEN);

#endif
    }
else
    {
    strncpy(callback_info[new_descriptor]->hostname,
            hs->h_name,MAXHOSTNAMELEN);
    }

callback_info[new_descriptor]->descriptor = new_descriptor;

do_callback(st_NEW_CONNECTION, new_descriptor);
return SUCCESS;
}

/****************************************************************************/
/* Function: close_connection           Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: descriptor to be closed                              */
/*                                                                          */
/* Return codes: NONE                                                       */
/*                                                                          */
/* Purpose: This function removes the descriptor from the global desc list  */
/*          as well as free'ing the socket structure associated with it,    */
/*          and closes the descriptor.                                      */
/*                                                                          */
/* Change History:                                                          */
/*                                                                          */
/* Charles Henrich  We now check to see if callback_info[descriptor] is     */
/* June 30th, 1993  NULL before going off and referencing it, and we also   */
/*                  check to see if descriptor is in the ready_descriptors  */
/*                  array, if it is we set it to -1 in said array so it     */
/*                  wont be processed back in st_mainloop.                  */
/*                                                                          */
/****************************************************************************/

static void close_connection(int descriptor)
{
int count;
int high_descriptor=0;

if(callback_info[descriptor] == NULL)
    {
#ifdef DEBUG
    printf("ST: Im closing a non-exitant connection here, this is nuts!\n");
#endif

    return;
    }

#ifdef DEBUG
printf("ST:Closing connection\n");
#endif

do_callback(st_CLOSE_CONNECTION,descriptor);

 /**************************************/
 /*                                    */
 /* Remove connection from descriptor  */
 /* list                               */
 /*                                    */
 /**************************************/

all_descriptors[descriptor] = 0;

 /**************************************/
 /*                                    */
 /* Remove this descriptor from the    */
 /* ready list, we dont want to be     */
 /* used if we didnt cause this close. */
 /*                                    */
 /**************************************/

for(count=0;count<nfds;count++)
    {
    if(ready_descriptors[count] == descriptor)
        {
        ready_descriptors[count] = -1;
        break;
        }
    }

 /**************************************/
 /*                                    */
 /* Compute and set the new high       */
 /* descriptor number;                 */
 /*                                    */
 /**************************************/

for(count=0;count<nfds;count++)
    {
    if(all_descriptors[count] == 1) high_descriptor=count;
    }

nfds=high_descriptor+1;

FD_CLR(descriptor, &readfds);         /* Remove descriptor from select list */

#ifdef DEBUG
printf("ST:freeing callback_info  descriptor %d\n",descriptor);
#endif

free(callback_info[descriptor]);

#ifdef DEBUG
printf("ST:After free\n");
#endif

callback_info[descriptor] = NULL;

#ifdef DEBUG
printf("ST:About to close descriptor\n");
#endif

close(descriptor);

#ifdef DEBUG
printf("ST:Descriptor closed\n");
#endif
}

/****************************************************************************/
/* Function: sock_write                 Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: descriptor and string to be written                  */
/*                                                                          */
/* Return codes: SUCCESS or FAILURE or MASSIVE_FAILURE                      */
/*                                                                          */
/* Purpose: This function blasts a string out the socket referenced by      */
/*          descriptor.                                                     */
/*                                                                          */
/****************************************************************************/

static int sock_write(int descriptor, char *string)
{
int ret;

#ifdef DEBUG
printf("ST:Sock write Desc:%d  String%s",descriptor, string);
#endif

#ifndef DES
ret=write(descriptor, string, strlen(string));    /* Write string to socket */
#else
if(G_encrypt == TRUE)
    {
    ret=des_enc_write(descriptor, string, strlen(string),
            G_dessched, G_deskey);
    }
else
    {
    ret=write(descriptor, string, strlen(string));/* Write string to socket */
    }
#endif

if(ret == -1)                                          /* We have an error! */
    {
    switch(errno)
        {
        case EMSGSIZE:                       /* All these are nasty errors  */
        case EINVAL:                         /* We should hang up if we get */
        case ENETUNREACH:                    /* any of them                 */
        case EHOSTUNREACH:
        case EPIPE:

#ifdef DEBUG
printf("ST:Returning massive failure\n");
#endif

            return MASSIVE_FAILURE;
            break;

        case EWOULDBLOCK:                  /* Recoverable error, dont close */
        default:                           /* the connection                */

#ifdef DEBUG
printf("Returning failure\n");
#endif

            return FAILURE;
            break;
        }
    }

do_callback(st_WRITE,descriptor);

return SUCCESS;
}

/****************************************************************************/
/* Function: sock_read                  Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: descriptor, buffer to read into, and size of buffer  */
/*                                                                          */
/* Return codes: SUCCESS or FAILURE or MASSIVE_FAILURE or PARTIAL_STRING    */
/*                                                                          */
/* Purpose: This function reads data that is waiting on a socket into the   */
/*          buffer pointed to by the second parameter.  It takes care not   */
/*          to overwrite any data in the buffer until it has a \n on the    */
/*          end of it.                                                      */
/*                                                                          */
/****************************************************************************/

static int sock_read(int descriptor, char *string, int bufsize)
{
int ret;
int count;
int bytes;
char buffer[BUF_SIZE];

if(bufsize > BUF_SIZE) bufsize = BUF_SIZE;

bufsize--;

#ifdef DEBUG
printf("ST:Sock Read\n");
#endif

#ifndef DES
if( (bytes = recv(descriptor, buffer, bufsize-strlen(string), MSG_PEEK)) < 1)
    {
    return MASSIVE_FAILURE;
    }

buffer[bytes] = '\0';

for(count=0;count < bytes;count++)
    {
    if(buffer[count] == '\n')
        {
        bytes=count+1;
        break;
        }
    }

ret=read(descriptor, buffer, bytes);
#else

if(G_encrypt == TRUE)
    {
    ret=des_enc_read(descriptor, buffer, bufsize-strlen(string), 
                     G_dessched, G_deskey);
    }
else
    {
    if((bytes = recv(descriptor, buffer, bufsize-strlen(string),MSG_PEEK)) < 1)
        {
        return MASSIVE_FAILURE;
        }

    buffer[bytes] = '\0';

    for(count=0;count < bytes;count++)
        {
        if(buffer[count] == '\n')
            {
            bytes=count+1;
            break;
            }
        }

    ret=read(descriptor, buffer, bytes);
    }
#endif

if(ret == 0)                            /* Read zero bytes, something wrong */
    {
    return MASSIVE_FAILURE;
    }

if(ret == -1)                                      /* Error occured in read */
    {
    switch(errno)
        {
        case ETIMEDOUT:                               /* Recoverable errors */
        case EWOULDBLOCK:
            return FAILURE;
            break;

        case ENETUNREACH:                           /* Unrecoverable errors */
        case EHOSTUNREACH:
        case ECONNRESET:
        case EMSGSIZE:
        default:
            return MASSIVE_FAILURE;
            break;
        }
    }

buffer[ret] = '\0';    /* Terminate buffer at the number of bytes read read */

strcat(string,buffer);      /* Add string to existing partially read string */

 /**************************************/
 /*                                    */
 /* Okay this is a hack, but it beats  */
 /* core dumping.  If our buffer is    */
 /* full, we're going to return it as  */
 /* a non-partial....                  */
 /*                                    */
 /**************************************/

if(strlen(string) >= bufsize-2) return SUCCESS;

 /**************************************/
 /*                                    */
 /* Check to see if the string is      */
 /* complete, I.e terminated with a    */
 /* \n                                 */
 /*                                    */
 /**************************************/

for(count=0;string[count];count++)
    {
    if(string[count] == '\n')
        {
        if(count > 0 && string[count-1] == '\r') 
            {
            string[count-1] = '\n';
            string[count] = '\0';
            }

        string[count+1] = '\0';
        return SUCCESS;
        }
    }

return PARTIAL_STRING;
}

/****************************************************************************/
/* Function: do_callback                Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: callback number, and descriptor                      */
/*                                                                          */
/* Return codes: NONE                                                       */
/*                                                                          */
/* Purpose: This function calls the function referenced by callback number  */
/*          for the socket structure that contains the descriptor           */
/*                                                                          */
/****************************************************************************/

static void do_callback(int callback, int descriptor)
{
#ifdef DEBUG
printf("ST:Doing callback %d, descriptor %d\n",callback,descriptor);
#endif

if(callback_list[callback] != NULL)
    {
    callback_info[descriptor]->callback=callback;
    callback_list[callback](callback_info[descriptor]);
    }

#ifdef DEBUG
printf("ST:Back from callback!\n");
#endif
}

/****************************************************************************/
/* Function: input_handler              Author: Charles Henrich             */
/*                                      Date  : January 11th 1993           */
/*                                                                          */
/* Calling Parameters: the descriptor                                       */
/*                                                                          */
/* Return codes: SUCCESS or FAILURE or MASSIVE_FAILURE or PARTIAL_STRING    */
/*                                                                          */
/* Purpose: This function retrieves data from a socket, calling the         */
/*          appropriate callback function if an entire string of data has   */
/*          been retrieved (string ends with \n)                            */
/*                                                                          */
/****************************************************************************/

static int input_handler(int descriptor)
{
int ret;
int numtok;

#ifdef DEBUG
printf("ST:Entered input_handler\n");
#endif

numtok = 0;

ret=sock_read(descriptor, callback_info[descriptor]->buffer, BUF_SIZE);

if(ret == MASSIVE_FAILURE) return MASSIVE_FAILURE; /* Ahh go away!! hangup! */

if(ret == FAILURE) return FAILURE;          /* Error in reading from socket */

if(ret == PARTIAL_STRING) return PARTIAL_STRING;         /* Need more data! */

#ifdef DEBUG
printf("sock_read returns %d\n",ret);
#endif

do_callback(st_READ, descriptor);

if(callback_info[descriptor] != NULL)
    {
    callback_info[descriptor]->buffer[0]='\0';
    }

return ret;
}

